/*  -------------------------------------------------------------

    Rossum's Playhouse  --  a client/server based robot simulator
    Rossum's Playhouse is also known under the name "RP1".
    Copyright (C) 1999  G.W. Lucas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------- */


package rp1demo.clientzero;
import  rp1.rossum.*;
import  rp1.rossum.event.*;
import  rp1.rossum.request.RsMotionRequest;

import java.io.*;

import java.awt.Color;

public class ClnMain extends RsClient implements RsRunnable
{
   protected RsMouseClickEvent    latestMouseClick;

   public RsBody               body;
   public RsEncoder         [] encoders;

   protected int               mouseMode;
   protected boolean           autoPilot;

   protected int               searchIndex;
   protected ClnPath           motionPath;
   protected int               motionPathIndex;
   protected ClnPath           searchPath;
   protected ClnPath           homePath;
   protected RsNavNode      [] searchNode;
   protected int               searchRelatedTimeoutIndex;

   protected double            linearSpeed;
   protected double            angularSpeed;

   protected ClnNavNet         navNet;

   protected String            targetName;
   protected double            xTarget;
   protected double            yTarget;

   protected RsPoint           drPoint;
   protected double            drTheta;


   //  phases of an autoPilot excursion

   static final int SEARCH_FOR_TARGET = 0;
   static final int WAIT_TO_APPROACH  = 1;
   static final int COMPUTE_APPROACH  = 2;
   static final int APPROACH_TARGET   = 3;
   static final int EXTINGUISH_TARGET = 4;
   static final int COMPUTE_TRIP_HOME = 5;
   static final int RETURN_HOME       = 6;
   static final int WAIT_FOR_RESTART  = 7;



   // implement a main to allow this class to serve as an Application -----
   public static void main(String args[]) throws IOException
   {
      ClnMain c = new ClnMain();
      c.initialize();   // throws IOException if unable to reach server
      c.run();
   }


   // constructors --------------------------------------------------------
   public ClnMain(){
      super();   // perform class initializations required for RsClient

      // the following factors are used to weight the search.
      // they are based on a rough estimate of the performance
      // of the simulacrum.  Linear speed is the straight-line
      // speed of the robot in meters/second.   angular speed is
      // equivalent to radians/second.  Let's assume that the robot wheels
      // spin fast enough to move it at 1 meter per second
      // With a wheel base of 0.18 meters, the angular
      // velocity would simply be the wheel velocity divided by the
      // pivot radius (since the pivot point is in the middle of
      // the axle, the pivot radius is simply half the wheel track, or 0.09
      // Because pivot maneuvers are error-prone, we specify a slower
      // speed for the turns.

      linearSpeed   = 1.0;
      angularSpeed  = 0.5*linearSpeed/0.09;
      drPoint    = new RsPoint(0.0, 0.0);
      drTheta = 0.0;
   }



   @Override
public void initialize() throws IOException {
      super.initialize();
      setupHandlers();
   }

   @Override
public void initialize(RsProperties rsp) throws IOException {
      super.initialize(rsp);
      setupHandlers();
   }

   private void setupHandlers(){

      // now perform initializations specific to this client ----------------

      // set initial values for state-related elements ---------------------
      latestMouseClick=null;

      try{
         mouseMode  = rsProperties.extractInt("mouseMode");
      }catch(RsPropertiesException rpe){}

      if(mouseMode<0 || mouseMode>1)
          mouseMode = 0;

      // create body and send its specification to server -------------------
      body = ClientZero.build(rsProperties);

      RsDifferentialSteering ds = (RsDifferentialSteering)body.wheelSystem;
      encoders = ds.addEncodersToDefaultWheels(256);
      sendBodySpecification(body);

      // register event handlers -----------------------------------------------
      addMouseClickEventHandler(new ClnMouseClickEventHandler(this));
      addPositionEventHandler(  new ClnPositionEventHandler(  this));

      addPlacementEventHandler(   new ClnPlacementEventHandler(   this));
      addMotionHaltedEventHandler(new ClnMotionHaltedEventHandler(this));
      addTargetSensorEventHandler(
            (RsBodyTargetSensor)(body.getPartByName("head")),
            new ClnTargetSensorEventHandler(this));
      addTargetSelectionEventHandler(new ClnTargetSelectionEventHandler(this));
      addTimeoutEventHandler(        new ClnTimeoutEventHandler(   this));
      addPlanEventHandler(           new ClnPlanEventHandler(      this));
      addEncoderStatusEventHandler(  new ClnEncoderStatusEventHandler(this));


      // for demo purposes we use the same object as an event handler
      // for all three contact sensors.  This isn't necessarily recommended
      // practice (nor is it deprecated practice), it's just intended
      // to show that it can be done.


      ClnContactSensorEventHandler cHandler = new ClnContactSensorEventHandler(this);
      addContactSensorEventHandler(
            (RsBodyContactSensor)(body.getPartByName("Contact 0")), cHandler);
      addContactSensorEventHandler(
            (RsBodyContactSensor)(body.getPartByName("Contact 1")), cHandler);
      addContactSensorEventHandler(
            (RsBodyContactSensor)(body.getPartByName("Contact 2")), cHandler);



      ClnRangeSensorEventHandler rHandler = new ClnRangeSensorEventHandler(this);
      addRangeSensorEventHandler(
            (RsBodyRangeSensor)(body.getPartByName("RangeSensor 0")), rHandler);
      addRangeSensorEventHandler(
            (RsBodyRangeSensor)(body.getPartByName("RangeSensor 1")), rHandler);
      addRangeSensorEventHandler(
            (RsBodyRangeSensor)(body.getPartByName("RangeSensor 2")), rHandler);
      addRangeSensorEventHandler(
            (RsBodyRangeSensor)(body.getPartByName("RangeSensor 3")), rHandler);



      sendPlanRequest();
      sendPlacementRequest("home");


      RsBodyPainter painter = (RsBodyPainter)(body.getPartByName("painter"));
      if(painter!=null)
         sendPainterActivationRequest(painter);

   }


   // PUBLIC interface methods ---------------------------------------
   // the following methods may be used by the GUI or other components

   public void setMouseMode(int mouseMode){
      this.mouseMode = mouseMode%2;
   }

   public int getMouseMode(){
      return mouseMode;
   }

   public void setAutoPilot(boolean status){
      if(status==true && navNet==null){
         log("Internal Error -- attempt to set Auto Pilot when there is no nav-net");
         return;
      }
      autoPilot=status;
      stopMotion();
      if(autoPilot)
         sendPlacementRequest();
      else
         sendAllTargetsSelectionRequest(true);
   }

   public void  stopMotion(){
      sendHaltRequest();
      latestMouseClick=null;
      motionPath=null;
      searchRelatedTimeoutIndex=0;
   }

   public void  sendPlacementRequest(){
      super.sendPlacementRequest("home");
      RsBodyPainter painter = (RsBodyPainter)(body.getPartByName("painter"));
      if(painter!=null && painter.getPainterActivationStatus()){
          painter.setPaintColor(new Color(192, 0, 128));
          sendPainterActivationRequest(painter);
      }
      latestMouseClick=null;
      motionPath=null;
   }


   public RsBody getBody(){
      return body;
   }

   protected boolean advanceMotionPath(){
      // return true if there was any moves remaining in motion,
      // otherwise returns false.

      if(motionPathIndex>=motionPath.nMove)
         return false;

      ClnMove  move = motionPath.move[motionPathIndex];
      motionPathIndex++;

      RsMotionRequest r;
      if(move.sAngular!=0.0){
         double xMove=Math.cos(move.sAngular);
         double yMove=Math.sin(move.sAngular);
         r = body.wheelSystem.getMotionRequestForPivot(xMove, yMove, 0.4);
         if(r!=null && r.duration>0){
            sendMotionRequest(r);
         }
         return true;
      }

      r = body.wheelSystem.getMotionRequest(move.sLinear, 0.0, 0.4);
      if(r!=null && r.duration>0)
         sendMotionRequest(r);
      return true;
   }


   public void initiateSearch(double x, double y, double orientation){
      searchIndex=SEARCH_FOR_TARGET;
      motionPathIndex=0;
      motionPath=searchPath;
      advanceMotionPath();
   }



   public void scheduleNextExcursion(double timeoutPeriod){
      searchIndex               = ClnMain.WAIT_FOR_RESTART;
      searchRelatedTimeoutIndex = sendTimeoutRequest(timeoutPeriod);
      motionPathIndex           = 0;
      motionPath                = null;
   }


   public void setFloorPlan(RsPlan floorPlan){
      if(floorPlan!=null && floorPlan.getNavLinkArray()!=null){
         // if we get a valid floorPlan which includes a nav-net
         // we can establish a ClnNavNet.  This test isn't perfect,
         // but it will do for this revision.
         navNet     = new ClnNavNet(floorPlan);
         searchPath = navNet.optimize(linearSpeed, angularSpeed);
         homePath   = new ClnPath();   // empty until needed
      }else{
         navNet     = null;
      }
   }

   public boolean isNavNetEstablished(){
      return navNet!=null;
   }
}
