/*  -------------------------------------------------------------

 Rossum's Playhouse  --  a client/server based robot simulator
 Rossum's Playhouse is also known under the name "RP1".
 Copyright (C) 1999  G.W. Lucas

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 ----------------------------------------------------------------- */

/*

 ClnNavNet.java


 About findPath

 The findPath method uses linearSpeed and angularSpeed as ways of
 weighting the score for each potential path it finds.
 These speeds should reflect the estimated performance of the
 simulated robot. Linear speed is the straight-line
 speed of the robot in meters/second.   Angular speed is
 equivalent to radians/second.  Let's assume that the robot wheels
 spin fast enough to move it at 1 meter per second
 With a wheel base of 0.18 meters, the angular
 velocity would simply be the wheel velocity divided by the
 pivot radius (since the pivot point is in the middle of
 the axel, the pivot radius is simply half the wheel base, or 0.09

 The speed values to not have to be perfect for the algorithm
 to work, only reasonable.  Without them, sub-optimal paths
 may be selected.  For example, if the angularSpeed value is
 zero, the algorithm might pick a path with a lot of turns
 rather than one which covers the same distance with fewer
 turns.

 */

package rp1demo.clientzero;

import rp1.rossum.*;

public class ClnNavNet {
	
	private RsPlan       floorPlan;
	private ClnPath      testPath;
	private ClnPath      bestPath;
	private RsNavNode    goal;
	private double       turnSpeed, lineSpeed;
	private double       bestScore;
	private ClnMove   [] testMove;
	private RsNavNode [] checkPointArray;
	private RsTarget  [] targetArray;
	private RsPlacement  home;
	private RsNavNode    homeNode;
	
	
	public ClnNavNet(RsPlan _floorPlan) {
		floorPlan = _floorPlan;

		// bestPath is used as a place to store the
		// best discovered path. testPath and testMove
		// are scratch storage used by the various methods
		// in this class.
		bestPath = new ClnPath();
		testPath = new ClnPath();
		testMove = new ClnMove[2];
		testMove[0] = new ClnMove();
		testMove[1] = new ClnMove();

		home = (RsPlacement) (floorPlan.getObjectByName("home"));
		homeNode = getNearestNode(home.x, home.y);
	}

	public RsPlan getPlan() {
		return floorPlan;
	}

	public RsNavNode getHomeNode() {
		return homeNode;
	}

	public RsNavLink getNearestLink(double x, double y, RsPoint p) {

		RsNavLink[] link = floorPlan.getNavLinkArray();
		if (link == null)
			return null;

		double min = 1.0e+32;
		int index = 0;
		RsVector a;
		RsSegment segment;
		double bx, by, t, tx, ty, dx, dy, m;

		for (int i = 0; i < link.length; i++) {
			segment = link[i].segment;
			a = segment.v;
			bx = x - segment.x;
			by = y - segment.y;

			// find x where b-x orthogonal to a, so (b-x) dot a = 0

			t = (a.x * bx + a.y * by) / segment.m / segment.m;
			if (t < 0)
				t = 0;
			else if (t > 1)
				t = 1;
			tx = t * a.x;
			ty = t * a.y;
			dx = bx - tx;
			dy = by - ty;
			m = dx * dx + dy * dy;
			if (m < min) {
				min = m;
				index = i;
				p.x = tx + segment.x;
				p.y = ty + segment.y;
			}
		}
		return link[index];
	}

	public RsNavNode getNearestNode(double x, double y) {
		RsNavNode[] node = floorPlan.getNavNodeArray();
		if (node == null)
			return null;

		RsNavNode nearest = null;
		double min = 1.0e+32;
		double m, dx, dy;

		for (int i = 0; i < node.length; i++) {
			dx = x - node[i].x;
			dy = y - node[i].y;
			m = dx * dx + dy * dy;
			if (m < min) {
				min = m;
				nearest = node[i];
			}
		}
		return nearest;
	}

	public RsNavNode getNearestCheckPoint(double x, double y) {
		RsNavNode nearest = null;
		double nearestD2 = 1.0e+32;
		;
		double dx, dy, d2;
		RsNavNode[] c;
		c = getCheckPointArray();
		for (int i = 0; i < c.length; i++) {
			dx = c[i].x - x;
			dy = c[i].y - y;
			d2 = dx * dx + dy * dy;
			if (d2 < nearestD2) {
				nearestD2 = d2;
				nearest = c[i];
			}
		}
		return nearest;
	}

	public RsNavNode[] getNavNodeEndpoints() {
		int i, n;

		RsNavNode[] nodeArray = floorPlan.getNavNodeArray();

		n = 0;
		for (i = 0; i < nodeArray.length; i++)
			if (nodeArray[i].link.length == 1)
				n++;

		if (n == 0)
			return null;

		RsNavNode[] a = new RsNavNode[n];
		n = 0;
		for (i = 0; i < nodeArray.length; i++)
			if (nodeArray[i].link.length == 1)
				a[n++] = nodeArray[i];

		return a;
	}

	public RsTarget[] getTargetArray() {

		if (targetArray == null) {

			if (floorPlan == null)
				return null;

			RsObject[] object = floorPlan.getObjectArray();
			if (object == null)
				return null;

			int i, n;
			n = 0;
			for (i = 0; i < object.length; i++) {
				if (object[i] instanceof RsTarget)
					n++;
			}

			if (n == 0)
				return null;

			targetArray = new RsTarget[n];
			n = 0;
			for (i = 0; i < object.length; i++) {
				if (object[i] instanceof RsTarget)
					targetArray[n++] = (RsTarget) object[i];
			}
		}

		return targetArray;
	}

	protected RsNavNode[] getCheckPointArray() {
		if (checkPointArray == null) {
			if (floorPlan == null)
				return null;
			RsNavNode[] node = floorPlan.getNavNodeArray();
			if (node == null)
				return null;

			int i, k;
			k = 0;
			for (i = 0; i < node.length; i++)
				if (node[i].getName().startsWith("CheckPoint"))
					k++;
			if (k == 0)
				return null;

			checkPointArray = new RsNavNode[k];
			k = 0;
			for (i = 0; i < node.length; i++)
				if (node[i].getName().startsWith("CheckPoint"))
					checkPointArray[k++] = node[i];
		}
		return checkPointArray;
	}

	protected ClnPath optimize(double linearSpeed, double angularSpeed) {

		if (home == null)
			return null;

		RsNavNode[] checkPoint = getCheckPointArray();
		if (checkPoint == null || checkPoint.length < 1)
			return null;

		RsNavNode goal;
		ClnPermute permute = new ClnPermute(checkPoint.length);
		ClnPath searchPath = new ClnPath();
		RsNavNode[] searchNode = new RsNavNode[checkPoint.length];
		ClnPath aPath = new ClnPath();
		ClnPath testPath = new ClnPath();
		double testScore;
		double probability;
		double bestScore = 1.0e+36;
		RsNavNode[] permutedNodes = new RsNavNode[checkPoint.length];
		double x, y, orientation;
		int i;

		for (int iPermute = 0; iPermute < permute.nFactorial; iPermute++) {
			x = home.x;
			y = home.y;
			orientation = home.orientation;
			RsNavNode start = homeNode;
			testScore = 0;
			testPath.clear();
			permute.getNextPermutation(checkPoint, permutedNodes);

			for (i = 0; i < permutedNodes.length; i++) {
				goal = permutedNodes[i];
				findPath(x, y, orientation, start, goal, angularSpeed,
						linearSpeed, aPath);

				// because the probability of the candle being in any
				// room is 1/numberOfRooms, each room in the sequence
				// has that much less a probability of being explored.
				// we weight the score according to that probabiltiy

				probability = 1.0 - i / (double) (permutedNodes.length);
				testScore += probability * aPath.duration;
				if (testScore > bestScore)
					break;
				testPath.append(aPath);
				aPath.computeFinalPosition(x, y, orientation);
				x = aPath.xFinal;
				y = aPath.yFinal;
				orientation = aPath.orientationFinal;
				start = goal;
			}

			if (testScore < bestScore) {
				bestScore = testScore;
				searchPath.copy(testPath);
				StringBuffer sb = new StringBuffer();
				for (i = 0; i < permutedNodes.length; i++) {
					searchNode[i] = permutedNodes[i];
					if (i > 0)
						sb.append(", ");
					sb.append(permutedNodes[i].getName());
				}
				System.err.println(">>> " + searchPath.duration + ">>> "
						+ sb.toString());
			}
		}

		searchPath.computeFinalPosition(home.x, home.y, home.orientation);
		x = searchPath.xFinal;
		y = searchPath.yFinal;
		orientation = searchPath.orientationFinal;
		findPath(x, y, orientation, searchNode[checkPoint.length - 1],
				homeNode, angularSpeed, linearSpeed, aPath);
		searchPath.append(aPath);
		searchPath.compact();
		return searchPath;
	}

	public void findPath(double x, double y, double theta, RsNavNode start,
			RsNavNode _goal, double _turnSpeed, double _lineSpeed,
			ClnPath _bestPath) {

		goal = _goal;
		turnSpeed = _turnSpeed;
		lineSpeed = _lineSpeed;
		bestPath = _bestPath;

		bestPath.clear();

		RsNavNode[] nArray = floorPlan.getNavNodeArray();

		for (int i = 0; i < nArray.length; i++)
			nArray[i].visited = false;

		bestScore = 1.0e+36;
		evalPath(x, y, theta, start, 0.0);
	}

	protected int calcMove(double x, double y, double theta, double xNext,
			double yNext, ClnMove[] move) {
		double d, dx, dy;
		double sinTheta, cosTheta, cosTurn, sinTurn, turn;

		dx = xNext - x;
		dy = yNext - y;
		d = Math.sqrt(dx * dx + dy * dy);
		if (d < 0.005) {
			move[0].sLinear = move[1].sLinear = 0;
			move[0].sAngular = move[1].sAngular = 0;
			move[0].duration = move[1].duration = 0;
			return 0;
		}

		dx /= d;
		dy /= d;
		sinTheta = Math.sin(theta);
		cosTheta = Math.cos(theta);
		sinTurn = dy * cosTheta - dx * sinTheta;
		cosTurn = dx * cosTheta + dy * sinTheta;
		turn = Math.atan2(sinTurn, cosTurn);
		if (Math.abs(turn) < 2.5 * Math.PI / 180.0) {
			move[0].sLinear = d;
			move[0].sAngular = 0;
			move[0].duration = d / lineSpeed;
			move[0].x = x + d * cosTheta;
			move[0].y = y + d * sinTheta;
			move[0].orientation = theta;
			move[1].sLinear = 0;
			move[1].sAngular = 0;
			move[1].duration = 0;
			return 1;
		} else {
			theta += turn;
			move[0].sAngular = turn;
			move[0].sLinear = 0;
			move[0].duration = Math.abs(turn) / turnSpeed;
			move[0].x = x;
			move[0].y = y;
			move[0].orientation = theta;
			move[1].sLinear = d;
			move[1].sAngular = 0;
			move[1].duration = d / lineSpeed;
			move[1].x = x + d * Math.cos(theta);
			move[1].y = y + d * Math.sin(theta);
			move[1].orientation = theta;
			return 2;
		}
	}

	private void evalPath(double x, double y, double theta, RsNavNode start,
			double score) {

		int i, n;
		RsNavLink link;
		RsNavNode node;
		double x1, y1, theta1, delta, newScore;

		start.visited = true; // make sure this is set to false when we
								// return.

		// first, see if any of the potential transfer nodes are
		// the goal... in that case, the evaluation is short-circuited
		// and we have the opportunity to quit early.
		for (i = 0; i < start.link.length; i++) {
			link = start.link[i];
			if (link.n0 == goal)
				node = link.n0;
			else if (link.n1 == goal)
				node = link.n1;
			else
				continue;
			// if we get here, the only valid move remaining is to our goal.
			// evaluate the path and see if it is the "best" found so far
			// no matter what, we are done evaluating this path and can return.
			n = calcMove(x, y, theta, node.x, node.y, testMove);
			newScore = score + testMove[0].duration + testMove[1].duration;
			if (newScore < bestScore) {
				bestScore = newScore;
				bestPath.copy(testPath);
				if (n > 0)
					bestPath.add(testMove[0], node.getName());
				if (n > 1)
					bestPath.add(testMove[1], node.getName());
			}
			start.visited = false;
			return;
		}

		for (i = 0; i < start.link.length; i++) {
			link = start.link[i];
			if (link.n0 == start)
				node = link.n1;
			else
				node = link.n0;
			if (node.visited)
				continue;
			n = calcMove(x, y, theta, node.x, node.y, testMove);
			if (n == 0) {
				evalPath(x, y, theta, node, score);
			} else {
				testPath.add(testMove[0], node.getName());
				if (n > 1) {
					testPath.add(testMove[1], node.getName());
					theta1 = theta + testMove[0].sAngular;
					delta = testMove[1].sLinear;
					newScore = score + testMove[0].duration
							+ testMove[1].duration;
				} else {
					theta1 = theta;
					delta = testMove[0].sLinear;
					newScore = score + testMove[0].duration;
				}
				x1 = x + Math.cos(theta1) * delta;
				y1 = y + Math.sin(theta1) * delta;
				evalPath(x1, y1, theta1, node, newScore);
				testPath.removeLast(n);
			}
		}
		start.visited = false;
	}

}
