/*  -------------------------------------------------------------

    Rossum's Playhouse  --  a client/server based robot simulator
    Rossum's Playhouse is also known under the name "RP1".
    Copyright (C) 1999  G.W. Lucas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------- */


package rp1demo.ackerman;
import  rp1.rossum.*;
import  rp1.rossum.request.RsActuatorControlRequest;

import java.io.*;

import java.awt.Color;

public class Main extends RsClient
{

   public RsBody   body;
   public double   linearSpeed;
   public double   angularSpeed;
   public double   turnSide;

   public static final double wheelBase      = 0.45;
   public static final double wheelTrack     = 0.3;
   public static final double stubAxleWidth  = 0.04;
   public static final double stubAxleLength = 0.02;
   public static final double tireRadius     = 0.08;
   public static final double tireWidth      = 0.04;
   public static final double maxInnerWheelSteeringAngle = Math.toRadians(75.0);
   public static final boolean dualSteering    = false; // true;
   // implement a main to allow this class to serve as an Application -----
   public static void main(String args[]) throws IOException
   {
      Main c = new Main();
      c.initialize();   // throws IOException if unable to reach server
      c.run();
   }


   // constructors --------------------------------------------------------
   public Main(){
      super();   // perform class initializations required for RsClient

      linearSpeed  = 0.5;
      turnSide     = 1;
   }



   @Override
public void initialize() throws IOException {
      super.initialize();
   }


   private void setupBody(){

      body = new RsBody("AckermanDemo");

      RsAckermanSteering wheelSystem = new
          RsAckermanSteering(
              wheelTrack,
              wheelBase,
              tireRadius,
              maxInnerWheelSteeringAngle,
              dualSteering);

      log("computed maxAbsSteeringAngle (deg): "+Math.toDegrees(wheelSystem.getMaxAbsSteeringAngle()));
      log("minimum turn radius (meters):       "+wheelSystem.getMinTurnRadius());

      wheelSystem.addDefaultWheels(tireRadius, tireWidth, stubAxleLength, stubAxleWidth);

      wheelSystem.setFillColor(Color.darkGray);
      wheelSystem.setLineColor(Color.darkGray);



      // TO DO:  we really should just add a static createNewRectangle method
      //         to RsBodyShape.

      // note that it is important that we build the robot with
      // interactive shapes (RsBodyShape) rather than simple RsBodyPart
      // objects (RsBodyShape extends RsBodyPart, adding interactivity).
      // set up a rectangular shape to represent chassis
      // note that the length of the chassis (wheelBase)
      // runs along the x axis, the width (wheelTrack) spreads out
      // across the y axis.
      double x0, y0, x1, y1;
      if(dualSteering){
         x0 = -wheelBase/2.0;
         x1 =  wheelBase/2.0;
      }else{
         x0 =  0;
         x1 =  wheelBase;
      }
      y0 = -wheelTrack/2.0;
      y1 =  wheelTrack/2.0;

      double [] p = new double[12];
      p[0] = x0;
      p[1] = y0;

      p[2] = x0;
      p[3] = y1;

      p[4] = x1;
      p[5] = y1;

      p[6] = x1+wheelTrack*0.25;   // make a "nose" point in front of chasis
      p[7] = (y0+y1)/2.0;

      p[8] = x1;
      p[9] = y0;

      p[10] = p[0];
      p[11] = p[1];

      RsBodyShape chassis = new RsBodyShape(p, 6);
      chassis.setName("chassis");
      chassis.setFillColor(Color.lightGray);
      chassis.setLineColor(Color.darkGray);

      body.addPart(chassis);
      body.addPart(wheelSystem);

   }

   public void sendNextMotionRequest(){

         RsAckermanSteering steering = (RsAckermanSteering)(body.wheelSystem);

         turnSide = -turnSide;
         double steeringAngle = turnSide*Math.toRadians(30.0);
         double driveWheelRV  = steering.mapVelocityToWheelRotationalVelocity(linearSpeed);

         double radius = steering.getTurnRadiusForSteeringAngle(steeringAngle);

         double duration = Math.abs((2.0*Math.PI*radius)/linearSpeed); // time = circumference/speed

         RsActuatorControlRequest acr =
               steering.getControlRequestForWheelRotationAndSteeringAngle(
                   driveWheelRV, steeringAngle);
         sendActuatorControlRequest(acr);
         sendTimeoutRequest(duration);
         log("Sent actuator control request");

   }


   @Override
public void run(){

       setupBody();
       sendBodySpecification(body);

       addPlacementEventHandler(new PlacementEventHandler(this));
       addTimeoutEventHandler(new TimeoutEventHandler(this));


       sendPlacementRequest("home");

       super.run();

   }


}
