/*  -------------------------------------------------------------

    Rossum's Playhouse  --  a client/server based robot simulator
    Rossum's Playhouse is also known under the name "RP1".
    Copyright (C) 1999  G.W. Lucas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------- */


package rp1demo.clientzero;
import  rp1.rossum.*;


/*

  Notes:

    Body parts are rendered in the order they are added to the body.
    So if two parts overlap, the last one added will be drawn on top
    of the first.   In this demo, the wheel system is defined early
    on, but it is not added until the very end. This ordering is used
    so that the wheels are visible. It is an aesthetic choice, not a functional one.

    See RsBodyPart documentation for a discussion of color selections and rendering..

*/

import java.awt.Color;

public class ClientZero {

   public static RsBody  build(RsProperties properties){
      RsBody body = new RsBody("ClientZero");

      // create a differential steering system with a track width of
      // 0.18, from wheel center to wheel center
      double trackWidth       = 0.18;
      double driveWheelRadius = 0.075;
      RsDifferentialSteering wheelSystem = new RsDifferentialSteering(trackWidth, driveWheelRadius);

      double tireWidth  = 0.03;
      wheelSystem.addDefaultWheels(driveWheelRadius, tireWidth);
      wheelSystem.addDefaultCasterWheel(-0.08, 0.0, 0.0375, 0.013);


      wheelSystem.setFillColor(Color.darkGray);
      wheelSystem.setLineColor(Color.darkGray);



      double circleRadius=0.15;  // 0.3 meters accross
      RsBodyCircle circle = new RsBodyCircle(0.0, 0.0, circleRadius);
      circle.setName("main circle");
      circle.setFillColor(Color.lightGray);
      circle.setLineColor(Color.darkGray);
      body.addPart(circle);




      int    i, j, k;
      double a;
      double [] c  = new double[56];
      c[0]=0;
      c[1]=0;
      for(i=0; i<3; i++){
         k=2;
         for(j=i*120; j<=(i+1)*120; j+=5){
            a=j*Math.PI/180.0;
            c[k++]=(circleRadius+0.001)*Math.cos(a);
            c[k++]=(circleRadius+0.001)*Math.sin(a);
         }

         RsBodyContactSensor contact = new RsBodyContactSensor(c, 26);
         contact.setName("Contact "+i);
         contact.setFillColor(null);
         contact.setLineColor(null);
         // when cold (no detection), sensor is not depicted
         // hot colors are still set and visible
         body.addPart(contact);
      }

      double x, y, vx, vy;
      a=Math.PI/4.0;
      for(i=0; i<4; i++){
         vx=Math.cos(a);
         vy=Math.sin(a);
         c[0]=(circleRadius-0.025)*vx;
         c[1]=(circleRadius-0.025)*vy;
         x=(circleRadius-0.001)*vx;
         y=(circleRadius-0.001)*vy;
         c[2]=x-vy*0.01;
         c[3]=y+vx*0.01;
         c[4]=x+vy*0.01;
         c[5]=y-vx*0.01;
         c[6]=c[0];
         c[7]=c[1];
         RsBodyRangeSensor rSensor= new RsBodyRangeSensor(
            c, 3,
            x, y,
            a,
            0.12, 3);
         rSensor.setName("RangeSensor "+i);
         rSensor.setFillColor(Color.blue);
         rSensor.setLineColor(Color.blue);
         body.addPart(rSensor);
         a+=Math.PI/2.0;
      }


      double  d1[] = {
         0.09, -0.025,
         0.16, -0.025,
         0.16,  0.025,
         0.09,   0.025
      };
      RsBodySensor s1 = new RsBodyTargetSensor(
         d1, 4,
         0.16, 0.0,
         0.0,
         45*Math.PI/180.0, 1.0,
         3, 3);
      s1.setName("head");
      s1.setFillColor(Color.blue);
      s1.setLineColor(Color.blue);
      body.addPart(s1);



      // if the properties specifies a painter width,
      // then add a painter to the robot body
      double painterWidth=0;
      try{
        painterWidth  = properties.extractDouble("painterWidth");
      }catch(RsPropertiesException rpe){
        painterWidth=0;
      }
      if(painterWidth>0){
         c[0]=-0.025;
         c[1]=-0.025;

         c[2]= 0.025;
         c[3]=-0.025;

         c[4]= 0.025;
         c[5]= 0.025;

         c[6]=-0.025;
         c[7]= 0.025;

         RsBodyPainter pnt = new RsBodyPainter(c, 4);
         pnt.setName("painter");
         pnt.setFillColor(Color.green);
         pnt.setLineColor(Color.green);
         pnt.setPaintColor(new Color(192, 0, 128));
         pnt.setTrailerPosition(0.0, 0.0);
         pnt.setTrailerWidth(painterWidth);
         body.addPart(pnt);
      }


      body.addPart(wheelSystem);
      return body;
   }
}

