
/* ---------------------------------------------

Non-Recursive Permutation Generator

This class allows you to build permutations
as an interative process.

The elements

  public final int  n;           number of elements in the list to be permuted
  public final int  nFactorial;  n!

The methods

  public ClnPermute(int n)

       creates an object of class ClnPermute, populating
       the public elements

  void getNextPermutation(Object []objects, Object []permutedObjects);
       permutes the objects array, storing the contents in the
       permutedObjects array.  The sequence of permutations
       repeats after n! calls.   Note that this algorithm does
       not create an "ordered set" of permutations.

Warning:

     Because the value of n! increases at a faster-than-exponential
     rate, it can become quite large very quickly.  The largest
     value of n that ClnPermute can process correctly is 12.
     13! is actually larger than the maximum value of a
     Java 8-byte integer.  It will still accept values of n larger
     than 12, it simply won't generate the entire permutation
     set.


The algorithm for this class was taken from
"Data Structures and Program Design" by Robert L. Kruse
where it was coded elegantly and compactly as a
recursive implementation.

The one problem with such a recursive approach is that the
permuted sequence is generated only at the deepest layer
of recursive calls.   Thus, the recursive calling stack
dictates the flow-of-control for the program and you must
code so that your processing is invoked from within the
recursive function (usually through the use of a callback
or other such method).

This implementation allows you to create new sequences
through a series of iterative calls.  Each time you call
the "excursion" function, a new sequence is generated.
For a set of N items, N! calls will result in N! unique
sequences (on the next call after that, it cycles back
to the beginning).

In terms of speed, we gain nothing by removing the recursion
here because we still invoke the getNextPermutation method
just as many times as a recursive method would be invoked.
Were the stack-based logic folded into a tight loop rather
than implmented as a method within a class, we would
see a definite improvement.


*/

package rp1demo.clientzero;

public class ClnPermute {

   public ClnPermute(int _n){
      n      = _n;
      pStack = new int[n+1];
      list   = new int[n+1];
      k      = 1;
      p      = 0;
      index  = 0;

      int itemp = 1;
      for(int i=2; i<=n && i<13; i++)
         itemp *= i;
      nFactorial = itemp;

      pStack[0]=pStack[1]=0;
   }


   public final int  n;
   public final int  nFactorial;

   private int p;
   private int k;
   private int index;

   private int [] pStack;
   private int [] list;



   private void sequence(Object []object, Object []buffer){
      int i=0, q=0;
      while(list[q]!=0){
        buffer[i++]=object[list[q]-1];
        q=list[q];
      }
   }

   public void getNextPermutation(Object []object, Object []buffer){

      index++;

      while(k<n){
         /* push on to stack */
         list[k]=list[p];
         list[p]=k;
         pStack[k++]=p;
         p=0;
      }


      list[k]=list[p];
      list[p]=k;
      sequence(object, buffer);

      list[p]=list[k];
      p=list[p];

      if(index==nFactorial){
        // we've just generated the last sequence...  reset the
        // sequence, clear the stack, etc.
        k=1;
        p=0;
        pStack[0]=pStack[1]=0;
        index=0;
        return;
      }

      /** time to exit, do we leave the stack as is or do we pop it? */
      while(p==0){
         k--;
         p=pStack[k];
         list[p]=list[k];
         p=list[p];
      }

      return;

   }
}
