/*  -------------------------------------------------------------

    Rossum's Playhouse  --  a client/server based robot simulator
    Rossum's Playhouse is also known under the name "RP1".
    Copyright (C) 1999  G.W. Lucas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------- */


/*
  This position event handler handles the position event which
  we requested when we got a mouse click.    If a position
  event comes in when we haven't received a mouse click,
  it is ignored (or passed on to any other mouse-click handlers
  that might be registered).
*/


package rp1demo.clientzero;
import  rp1.rossum.*;
import  rp1.rossum.event.RsPositionEventHandler;
import  rp1.rossum.event.RsPositionEvent;
import  rp1.rossum.request.RsMotionRequest;

import java.lang.Math;

public class ClnPositionEventHandler implements RsPositionEventHandler {

   private ClnMain   client;

   // constructors -----------------------------------------------
   public ClnPositionEventHandler(ClnMain client){
      this.client = client;
   }

   // methods related to interface -------------------------------
   public void process(RsPositionEvent event){

      if(!client.autoPilot){

         if(client.latestMouseClick==null)
            return;

         client.body.setMotion(
            new RsMotionNull(
               event.simTime,
               event.orientation,
               event.x,
               event.y));

         double dx, dy, x, y;
         double sinTheta, cosTheta;

         dx = client.latestMouseClick.x - event.x;
         dy = client.latestMouseClick.y - event.y;
         cosTheta = Math.cos(event.orientation);
         sinTheta = Math.sin(event.orientation);
         // rotate delta vector by -event.orientation
         // to place it in the robot's frame of reference.
         x = dx*cosTheta + dy*sinTheta;
         y = dy*cosTheta - dx*sinTheta;

         RsMotionRequest r=null;

         if(client.latestMouseClick.button==1){
            r = client.body.wheelSystem.getMotionRequestForPivot(x, y, 0.3);
            if(r!=null)
               client.sendMotionRequest(r);
         }else{
            if(client.mouseMode==0){
               // note that we ignore the y-coordinate... this will cause the
               // the motion request to be computed as a straight line
               r = client.body.wheelSystem.getMotionRequest(x, 0, 0.3);
               if(r!=null)
                  client.sendMotionRequest(r);
            }else if(client.mouseMode==1){
               // here we use the y coordinate to develop a curved path
               r = client.body.wheelSystem.getMotionRequest(x, y, 0.3);
               if(r!=null)
                  client.sendMotionRequest(r);
            }
         }

         client.latestMouseClick=null;

      }else{ // autoPilot is true
         if(client.searchIndex==ClnMain.COMPUTE_APPROACH){
            ClnMove [] move = new ClnMove[2];
            move[0] = new ClnMove();
            move[1] = new ClnMove();
            client.motionPath = new ClnPath();

            int n = client.navNet.calcMove(
                 event.x, event.y, event.orientation,
                 client.xTarget, client.yTarget, move);

            if(n==2)
               client.motionPath.add(move[0]);
            if(n>0){
               ClnMove m = move[n-1];
               // by the contest rules, we need to approach to within
               // 12 inches (0.3 meters) of target.
               if(m.sLinear>0.3){
                  m.sLinear-=0.3;
                  client.motionPath.add(m);
               }
            }

            client.log("Approach initiated with nMove="+client.motionPath.nMove);
            if(client.motionPath.nMove>0){
               client.searchIndex=ClnMain.APPROACH_TARGET;
               client.motionPathIndex=0;
               client.advanceMotionPath();
            }else{
               // we are already in position to extinguish the target,
               // there is no need for further motion.
               client.searchRelatedTimeoutIndex = client.sendTimeoutRequest(3.000);
               client.searchIndex=ClnMain.EXTINGUISH_TARGET;
            }
         }else if (client.searchIndex==ClnMain.COMPUTE_TRIP_HOME){
            RsNavNode startNode = client.navNet.getNearestCheckPoint(event.x, event.y);
            client.motionPath=client.homePath;
            client.navNet.findPath(
               event.x, event.y, event.orientation,
               startNode, client.navNet.getHomeNode(),
               client.linearSpeed, client.angularSpeed,
               client.motionPath);
            client.log("Trip home computed with nMove="+client.motionPath.nMove);
            client.searchIndex=ClnMain.RETURN_HOME;
            client.motionPathIndex=0;
            client.advanceMotionPath();
         }
      }
   }
}
