/*  -------------------------------------------------------------

    Rossum's Playhouse  --  a client/server based robot simulator
    Rossum's Playhouse is also known under the name "RP1".
    Copyright (C) 1999  G.W. Lucas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------- */


package rp1demo.drunk;
import rp1.rossum.*;
import rp1.rossum.request.RsMotionRequest;
import rp1demo.clientzero.ClientZero;

import java.io.*;
import java.util.*;


//
// note that by extending RsClient, this class inherits
// a complete implementation of the RsRunnable interface.
// it uses three of the RsClient methods  -- setInputOutputStreams,
// setLogger, and initialize -- directly, and uses the run method
// by overriding it, but then invoking it as super.run() within
// its own run method.
//


public class DrunkMain extends RsClient
{

   public  RsBody            body;
   public  Random            random;

   // implement a main to allow this class to serve as an Application -----
   public static void main(String args[]) throws IOException
   {
      DrunkMain c = new DrunkMain();
      c.initialize();   // throws IOException if unable to reach server
      c.run();
   }


   public DrunkMain(){
         super();
   }


   @Override
public void run(){

      body = ClientZero.build(rsProperties);
      sendBodySpecification(body);

      // see if the rossum.ini file contains a specification
      // for the number of movements the drunk robot is to
      // perform in its random walk
      int nMovesForTest = 1000;
      try{
         nMovesForTest  = rsProperties.extractInt("nMovesForTest");
      }catch(RsPropertiesException rpe){}

      addPlacementEventHandler( new DrunkPlacementEventHandler( this));
      addMotionHaltedEventHandler(new DrunkMotionHaltedEventHandler(this, nMovesForTest));
      addMotionStartedEventHandler(new DrunkMotionStartedEventHandler(this));


      try{
         int seed = rsProperties.extractInt("seed");
         log("DrunkMain establishing random series with seed "+seed);
         random = new Random(seed);
      }catch(RsPropertiesException rpe){
         random = new Random();
      }

      log("Beginning series of "+nMovesForTest+" movements");
      sendPlacementRequest("home");

      RsBodyPart painter = body.getPartByName("painter");
      if(painter!=null)
         sendPainterActivationRequest((RsBodyPainter)painter);

      super.run();
   }



   protected double nextMove(){

      RsMotionRequest  r;
      double           x, y;
      boolean          pivot;

      do{
         pivot = (random.nextDouble()>0.75);
         if(pivot){
            x=2*random.nextDouble()-1.0;
            y=2*random.nextDouble()-1.0;
            r = body.wheelSystem.getMotionRequestForPivot(x, y, 0.2);
         }else{
            x=(2*random.nextDouble()-1.0)*0.25;
            y=(2*random.nextDouble()-1.0)*0.4;
            r = body.wheelSystem.getMotionRequest(x, y, 0.2);
         }
      }while(r==null);

      verbose("sending motion request:\n   linearVelocity "+r.linearVelocity+
                                     "\n   rotationalVelocity: "+r.rotationalVelocity+
                                     "\n   duration: "+r.duration);
      sendMotionRequest(r);
      return r.duration;
   }


}

