/*  -------------------------------------------------------------

    Rossum's Playhouse  --  a client/server based robot simulator
    Rossum's Playhouse is also known under the name "RP1".
    Copyright (C) 1999  G.W. Lucas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------- */



/*

   Elements related to line-tracker logic

      Dynamic values which are modified during the excursion

         wL, wR     wheel speeds in m/s, left and right wheel respectively


         leftHot, rightHot  boolean sensor values (true <==> detection)

         leftRegion, rightRegion   artifical values with no real-world
                                   equivalent.  used to control simulation.


      Parameters set at beginning of the excursion.  These reflect
      estimates of the parameters or performance limits of the robot
      being studied, or specifications for its behavior.

         aP, aN     wheel acceleration in m/s^2,  set before beginning of
                    the excursion,

         wMin       minimum speed allowed during excusion.  The logic
                    implemented here permits negative values (for very tight
                    turns), but that feature may be of limited value

         wMax       maximum speed, wMax>0 and wMax>wMin

         wMinMean   this is the minimum mean speed to which we
                    allow the robot to drop during the excursion.

         sampleInterval   Interval at which the robot can query sensors.
                          Sample rate, in Hertz, is 1/sampleInterval.

*/



package rp1demo.linetracker;
import  rp1.rossum.*;
import  rp1.rossum.request.RsActuatorControlRequest;


import java.io.*;
import java.text.*;


public class LtMain implements RsRunnable
{

   public static void main(String [] args)  throws IOException
   {
      LtMain lt = new LtMain();
      lt.initialize();
      lt.run();
   }

   public LtMain(){

      rsClient = new RsClient();

      // set initial states
      wL   =  0.0;
      wR   =  0.0;
      sLine = 0.0;
      sTurn = 0.0;
      wMin =  0.01;
      wMax =  0.075;
      sThreshold = (wMin+wMax)*0.5;
      acceleration = 0.2;
      braking      = 0.2;
      sampleInterval = 0.05;   //  1/sampleInterval = sample rate in Hz

      wLprior = 0.0;
      wRprior = 0.0;
      leftHotPrior  = false;
      rightHotPrior = false;

      wA = acceleration*sampleInterval;
      wB = braking*sampleInterval;

      nf = NumberFormat.getInstance();
      nf.setMaximumFractionDigits(4);
      nf.setMinimumFractionDigits(4);
      nf.setMaximumIntegerDigits(2);
      nf.setMinimumIntegerDigits(1);
   }


   public void setInputOutputStreams(InputStream input, OutputStream output){
      rsClient.setInputOutputStreams(input, output);
   }


   public void setLogger(RsLogInterface logger){
      rsClient.setLogger(logger);
   }

   public void initialize() throws IOException {
      RsProperties rsp = new RsProperties();
      rsp.loadFromResource(this, "linetracker.ini");
      initialize(rsp);
   }


   /** @deprecated
     *  Originally, the initialize method in LtMain was intended to work with
     *  the discussion in the UsersGuide to demonstrate flexibility in establishing
     *  logging and connections; instead, it just turned out to be confusing
     *  and needlessly complicated. Use one of the other demonstration clients
     *  as an example of how to initialize a RP1 session.
     */

   @Deprecated
public void initialize(RsProperties rsp) throws IOException {

      if(!rsClient.isLoggerSet()){
         rsClient.initiateLogger(rsp);
      }
      if(!rsClient.isInputSet()){
         rsClient.initiateConnection(rsp);
      }

      body = LtBody.build();
      trackWidth = body.wheelSystem.getTrackWidth();

      rsClient.sendBodySpecification(body);

      System.out.println("Registering event handlers");

      rsClient.addPaintSensorEventHandler(
            (RsBodyPaintSensor)(body.getPartByName("Left Paint Sensor")),
            new LtPaintSensorEventHandler(this, 1));
      rsClient.addPaintSensorEventHandler(
            (RsBodyPaintSensor)(body.getPartByName("Right Paint Sensor")),
            new LtPaintSensorEventHandler(this, -1));


      rsClient.addHeartbeatEventHandler(new LtHeartbeatEventHandler(this));
      rsClient.addPlacementEventHandler(new LtPlacementEventHandler(this));
      rsClient.sendPlacementRequest("startLine");

   }

   public void run(){
      rsClient.run();
   }



   protected void computeAndSetMotion(){



      double    w;
      boolean   swap;



      if(leftHotPrior && !leftHot && !rightHot || rightHotPrior && !leftHot && !rightHot){
         // we gone from hot to cold.  on the assumption that we have
         // over-corrected the steering, we slip it back to a previous setting
         wL=wLprior;
         wR=wRprior;
      }else{
         wLprior = wL;
         wRprior = wR;
         if(leftHot || rightHot){
           if(sLine>sThreshold){
              sLine-=wB;
              if(sLine<sThreshold)
                sLine=sThreshold;
           }
           aTurn = 2*wA/trackWidth;

           if(leftHot)
              sTurn+=aTurn;
           else
              sTurn-=aTurn;

           wR=(2*sLine+trackWidth*sTurn)/2.0;
           wL=(2*sLine-trackWidth*sTurn)/2.0;

           if(wR>wMax)
              wR=wMax;
           if(wR<wMin)
              wR=wMin;
           if(wL>wMax)
              wL=wMax;
           if(wL<wMin)
              wL=wMin;

         }else{
            // both sensors are cold
            if(wL==wR){
               w=wL+wA/2.0;
               if(w<wMax)
                  wL=wR=w;
            }else{
               swap = false;
               if(sTurn<0){
                   swap = true;
                   w    = wL;
                   wL   = wR;
                   wR   = w;
               }
               // increase speed, maintaining proportion of two wheel rates
               w = wR+wA/2.0;
               if(w<=wMax){
                 wL *= (w/wR);
                 wR  = w;
               }
               if(swap){
                   w    = wL;
                   wL   = wR;
                   wR   = w;
               }
            }
         }
      }

      sTurn = (wR-wL)/trackWidth;
      sLine = (wR+wL)/2.0;
      leftHotPrior  = leftHot;
      rightHotPrior = rightHot;

      RsDifferentialSteering ds = (RsDifferentialSteering)body.wheelSystem;
      double left  = wL/ds.getDriveWheelRadius();
      double right = wR/ds.getDriveWheelRadius();
      RsActuatorControlRequest acr = ds.getControlRequestForWheelRotation(left, right);
      rsClient.sendActuatorControlRequest(acr);

   }



   // elements related to problem setup or diagnostics

   protected double acceleration;
   protected double braking;
   protected double trackWidth;
   protected double sLine, sTurn;



   // constant elements used in line-tracker logic
   protected double wA;
   protected double wB;
   protected double aTurn;
   protected double wMin;
   protected double wMax;
   protected double sThreshold;
   protected double sampleInterval;

   // dynamic elements used in line-tracker logic
   protected double wL;
   protected double wR;
   protected double wLprior;
   protected double wRprior;

   // state information for sensors (used in line-tracker logic)
   protected boolean           leftHot, rightHot;
   protected boolean           leftHotPrior, rightHotPrior;




   protected RsClient          rsClient;
   public RsBody               body;
   long                        time0, time1;   // used for time-trials
   protected boolean           verbosity;
   protected int               leftRegion, rightRegion;
   private NumberFormat        nf;
}
